#!/usr/bin/env python3
"""This module manages the final merging of results from each single module."""

import pandas as pd
from itertools import chain
import numbers


def multiply_frames_by(frames: list, weights: list):
    """This function multiplies 'Score' column in each DataFrame.

    Each DataFrame's 'Score' column is multiplied by corresponding value in weights list.

    Args:
        frames (list): List of dataframes to be multiplied.
        weights (list): Corresponding list of weights to be used in multiplication.
    Raises:
        TypeError: When given obj is of wrong type.
    Returns:
        (list) List of multiplied DataFrames.
    """
    if not isinstance(frames, list):
        raise TypeError("Given obj is not a list.")
    if not isinstance(weights, list):
        raise TypeError("Given obj is not a list.")
    if not all(isinstance(x, pd.DataFrame) for x in frames):
        raise TypeError("Some of the DataFrames are not DataFrame objects.")
    if not all(isinstance(x, numbers.Number) for x in weights):
        raise TypeError("Some of the weights are not numbers.")

    results = []
    for weight, dframe in zip(weights, frames):
        dframe['Score'] = dframe['Score'] * weight
        results.append(dframe)
    return results


def main(args):
    """Main function for running module.
    Args:
        args (argparse obj): Arguments from right subparser.
    """
    if len(args.files) != len(args.weights):
        raise ValueError("Number of files and weights is different.")
    if sum(args.weights) != 1:
        raise ValueError("Weights don't sum up to 1.")
    frames = list(map(pd.read_csv, args.files))
    frames = multiply_frames_by(frames, args.weights)

    final = pd.DataFrame(columns=['Virus', 'Host', 'Score'])
    final['Virus'] = list(chain(*list(map(lambda df: df['Virus'].tolist(), frames))))
    final['Host'] = list(chain(*list(map(lambda df: df['Host'].tolist(), frames))))
    final['Score'] = list(chain(*list(map(lambda df: df['Score'].tolist(), frames))))
    final['Score'] = final['Score'].apply(pd.to_numeric)

    final = final.groupby(["Virus", "Host"]).sum().reset_index()
    idx = final.groupby(['Virus'])['Score'].idxmax()
    final = final.loc[idx].sort_values('Score', ascending=False).reset_index(drop=True)

    final['Score'] = final['Score'].round(decimals=2)
    final.to_csv(args.output, index=False)



import json
from shutil import which

LOGO = r"""
      _   _    ____  ____  _____ ____  
     | | / \  / ___||  _ \| ____|  _ \ 
  _  | |/ _ \ \___ \| |_) |  _| | |_) |
 | |_| / ___ \ ___) |  __/| |___|  _ < 
  \___/_/   \_\____/|_|   |_____|_| \_\    

  ========================================

            |-----------------|
             \---------------/
              ~-_---------_-~
                 ~-_---_-~
                    ~-_
                 _-~---~-_
              _-~---------~-_
             /---------------\
            |-----------------|                               
"""

TYPES = ('fasta', 'fa', 'fna')


def chunks(lst: list, n: int):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def parse_config(config_path: str, default_config: dict) -> dict:
    try:
        with open(config_path, 'r') as fh:
            return json.load(fh)
    except (FileNotFoundError, ValueError, TypeError) as e:
        return default_config


def perform_tool_check(name: str):
    """Check whether `name` is on PATH and marked as executable.

    `name` is a tool which JASPER will use, for instance `blastn`.

    Args:
        name (str): Name of tool to perform check on.
    Returns:
        (bool): Value indicating whether tool exists in PATH or not.
    """
    return which(name) is not None



#!/usr/bin/env python3
"""This module manages the blast operations.

This module uses Biopython and NCBI-Blast+ for making local blast queries and parsing the output.

More about it:
    1. https://biopython.org/

    2. https://biopython.org/docs/dev/api/Bio.Blast.Applications.html

    3. https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2447716/
"""
from __future__ import annotations

import io
import os
import subprocess
import pandas as pd
from pathlib import Path
from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.Blast.Applications import NcbimakeblastdbCommandline

from . import utils


class Database:
    def __init__(self, source_dir: Path, name: str) -> None:
        """Inits database with given name
        Args:
            source_dir (Path): Path to directory containing host files.
            name (str): Name of database.
        Raises:
            TypeError: When source_dir or path is not valid obj type
            FileNotFoundError: When given path does not exist or is not a directory.
        """
        if not isinstance(source_dir, Path):
            raise TypeError("Given object is not a Path object")
        if not isinstance(name, str):
            raise TypeError(f"Name expected to be str")
        if not Path(source_dir).exists():
            raise FileNotFoundError("Given path does not exist.")
        if not Path(source_dir).is_dir():
            raise FileNotFoundError("Given path is not a directory")
        self.source_dir = Path(source_dir)
        self.name = name

    def _repair_fasta(self, file: Path):
        """This function reads single fasta file and returns its content.

        Raises:
            TypeError: When given file is of wrong type.
            FileNotFoundError: When given path is not file.
        Returns:
            (str): Repaired content of the file.
        """
        if not isinstance(file, Path):
            raise TypeError("File is not a Path object.")
        if not file.is_file():
            raise FileNotFoundError("Given path is not a file.")

        repaired_content: list = []
        contig: int = 1
        with open(file, 'r') as fh:
            for line in fh:
                if line.startswith(">"):
                    seq_id = f">{file.stem}|{contig}\n"  # Read header
                    repaired_content.append(seq_id)
                    contig += 1
                else:
                    repaired_content.append(line)
        return "".join(repaired_content)

    def _aggregate(self, directory: Path, outfile: Path):
        """This function aggregates every file in source directory.

        Aggregated file consists of every repaired file in source dir.

        Args:
             directory (Path): Path to directory containing files.
             outfile (Path): Name of the aggregation file to be created.
        Raises:
            TypeError: When given object is of wrong type.
            FileNotFoundError: When given directory does not exist.
            ValueError: When output file is empty.
        """
        if not isinstance(directory, Path):
            raise TypeError("Given object is not Path object.")
        if not isinstance(outfile, Path):
            raise TypeError("Given object is not Path object.")
        if not directory.exists():
            raise FileNotFoundError('Given path does not exist.')
        if not directory.is_dir():
            raise FileNotFoundError('Given path is not a directory.')

        if outfile.exists():
            outfile.unlink()

        for source_file in directory.iterdir():
            if not source_file.name.endswith(utils.TYPES):
                continue
            with open(outfile, 'a+') as fh:
                fh.write(self._repair_fasta(source_file))

        if os.path.getsize(outfile) == 0:  # size == 0
            raise ValueError("Blast input file is empty. Check your input.")

    def create(self) -> tuple:
        """Function for making local blast database.

        This function creates database from files found in source_dir.

        Returns:
            tuple(Database, str): Database object, output from makeblastdb.
        Creates:
            (*.nhr, *.nin, *.nsq): Created database's files in LMBD format.
        Raises:
            SubprocessError: When makeblastdb returns error or when input file does not exist.
        """

        self._aggregate(self.source_dir, Path("blast_input.fasta"))
        try:
            cmd = NcbimakeblastdbCommandline(input_file="blast_input.fasta",
                                             dbtype="nucl",
                                             title=self.name,
                                             out=self.name)
            makeblastdb_output = subprocess.run(str(cmd), capture_output=True, shell=True)
            if makeblastdb_output.stderr:
                if not Path("blast_input.fasta").exists():
                    raise subprocess.SubprocessError(
                        "Makeblastdb returned error. Input file for Makeblastdb does not exist. Check your input.",
                        str(cmd))
                else:
                    raise subprocess.SubprocessError("Makeblastdb returned error. Check your input.",
                                                     makeblastdb_output.stderr.decode())
        except Exception:
            raise
        finally:
            if Path("blast_input.fasta").exists():
                Path("blast_input.fasta").unlink()
            return self, makeblastdb_output.stdout.decode()

    def query(self, query_dir: Path, config: dict, blast_format: str, headers: tuple) -> pd.DataFrame:
        """This function queries content of the directory to created database

        Args:
            query_dir (Path): Directory containing query files.
            config (dict): Blast configuration dict.
            blast_format (str): Blast output format.
            headers ( tuple(*str) ): Headers matching blast output for final DataFrame.
        Raises:
            TypeError: When given obj is of wrong type.
            FileNotFoundError: When given path does not exist or when given path is not a directory.
            ValueError: When forbidden blast option was provided.
        Returns:
            (pd.DataFrame): Pandas DataFrame containing query results.
        """

        if not isinstance(query_dir, Path):
            raise TypeError("Given object is not Path object")
        if not query_dir.exists():
            raise FileNotFoundError("Given path does not exist")
        if not query_dir.is_dir():
            raise FileNotFoundError("Given path is not directory")

        if not isinstance(config, dict):
            raise TypeError("Config file is not a dict object")
        if any(kwarg in ('query', 'db', 'outfmt', 'max_target_seqs', 'num_alignments') for kwarg in config.keys()):
            used = filter(lambda k: k in config.keys(), ('query', 'db', 'outfmt', 'max_target_seqs', 'num_alignments'))
            raise ValueError("Given kwargs are not valid in terms of blast usage", list(used))

        self._aggregate(query_dir, Path("blast_query.fasta"))
        try:
            cmd = NcbiblastnCommandline(query="blast_query.fasta",
                                        db=f"{self.name}",
                                        outfmt=blast_format,
                                        max_target_seqs=1,
                                        **config)
            blastn_output = subprocess.run(str(cmd), capture_output=True, shell=True)

            # Error only occurs if it's not this stupid warning.
            if blastn_output.stderr and "Examining 5 or more matches" not in blastn_output.stderr.decode():
                if not Path("blast_query.fasta").exists():
                    raise subprocess.SubprocessError("Blastn returned error. Input file for Blastn does not exist. Check your input.", str(cmd))
                else:
                    raise subprocess.SubprocessError("Blastn returned error. Check your input.", blastn_output.stderr.decode())
        except Exception:
            raise
        finally:
            if Path("blast_query.fasta").exists():
                Path("blast_query.fasta").unlink()
        results_df: pd.DataFrame = pd.read_csv(io.StringIO(blastn_output.stdout.decode()), header=None, names=headers)
        return results_df

    def clear_files(self):
        """Function clears database files if any present."""
        for file in Path(".").iterdir():
            if file.stem == self.name and file.suffix in (".nhr", ".nin", ".nsq"):
                file.unlink()


def main(args):
    """Main function for running module.
    Args:
        args (argparse obj): Arguments from right subparser.
    """
    print(utils.LOGO)

    # Default config if not specified
    args.blastn_config = utils.parse_config(args.blastn_config, {
        "task": "blastn",
        "num_threads": os.cpu_count(),
    })
    print("Starting analysis...")
    if args.create_db_name:
        print("Aggregating files, creating database...")
        db, db_output = Database(Path(args.host_dir), args.create_db_name).create()
        print(db_output)
    else:
        db = Database(Path('.'), args.use_db_name)
    print("Quering...")
    query_df = db.query(Path(args.virus_dir),
                        blast_format="10 qseqid sseqid score",
                        headers=("Virus", "Host", "Score"),
                        config=args.blastn_config)
    if args.clear_after:
        db.clear_files()

    query_df['Score'] = query_df['Score'].apply(pd.to_numeric)
    print("Blastn results (genome-genome query): ", query_df, sep='\n')

    query_df.to_csv(args.output_file, index=False)
    print("Saved files to", args.output_file)


class Expando(object):
    pass


if __name__ == "__main__":
    args = Expando()
    args.virus_dir = Path("example_data/virus")
    args.host_dir = Path("example_data/host")
    args.create_db_name = "host_db"
    args.blastn_config = ""
    args.clear_after = True
    args.output_file = "blast_results.csv"
    main(args)



#!/usr/bin/env python3
"""This module manages the crispr parsing and analysis.

This module uses Biopython and Piler-CR for identification of CRISPR spacers and their analysis.

More about it:
    1. https://biopython.org/

    2. https://biopython.org/docs/dev/api/Bio.Blast.Applications.html

    3. http://www.ncbi.nlm.nih.gov/pubmed/17239253
"""

from __future__ import annotations

import os
import shutil
import subprocess
import pandas as pd
from pathlib import Path

from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from Bio import SeqIO

from . import blast
from . import utils


class CrisprFinder(blast.Database):
    def __init__(self, source_dir: Path, name: str) -> None:
        """Inits obj with args"""
        super(CrisprFinder, self).__init__(source_dir, name)

    def _make_output_dir(self, directory: Path):
        """This function creates output dir for PILER-CR

        Args:
            directory (Path): Directory to be created.
        Raises:
            TypeError: When given path is not Path obj.
        """
        if not isinstance(directory, Path):
            raise TypeError("Given object is not Path object.")

        if not directory.exists():
            directory.mkdir()

    def retrieve_spacers(self) -> CrisprFinder:
        """This function retrieves spacers from files in source directory

        This method uses blast module functionality.
        It runs PILER-CR and parses output for each file in source dir.
        Then it creates a virus database and uses retrieved spacers as a blastn-short query.
        Resulting dataframe is populated with blast query results.
        """
        res_dir = Path("crispr_spacers")
        self._make_output_dir(res_dir)

        for host in self.source_dir.iterdir():
            if not host.name.endswith(utils.TYPES):
                continue

            repaired_file = Path(f"{host.stem}.repaired")
            with open(repaired_file, "w+") as repaired_fh:
                repaired_fh.write(self._repair_fasta(host))

            piler_file = res_dir / Path(f"{host.stem}.piler")  # Output file for PILERCR.
            self.find_crispr_spacers(repaired_file, piler_file)  # Find crispr spacers for given file.
            repaired_file.unlink()  # Remove temp file for repaired seq

            # Process the PILERCR output
            with open(piler_file, 'r') as res_fh:
                # Read the content of PILERCR file.
                piler_content = res_fh.read()
                # If spacer not found then continue
                if "DETAIL REPORT" not in piler_content:
                    piler_file.unlink()
                    continue
                # Get only spacers region
                piler_content = piler_content.split("DETAIL REPORT")[1].split("SUMMARY BY SIMILARITY")[0]
            # Delete PILERCR output file
            piler_file.unlink()

            for line in piler_content.splitlines():
                if line.startswith(">"):
                    name = f"{line.lstrip('>').split(' ')[0]}"
                try:
                    line = list(filter(lambda x: x, line.split(" ")))  # Filter non empty strings
                    if len(line) == 7 and any(base in line[6] for base in "ATGC"):
                        # If line had 7 fields and any of 'ATGC' in 7th field
                        seq = SeqRecord(Seq(line[6]), id=name, description="", name=name)  # Create seq
                        spacers_file = res_dir / Path(f"{name}.fasta")  # Result file for spacers
                        with open(spacers_file, 'a+') as final_fh:  # Append seq to file
                            SeqIO.write(seq, final_fh, 'fasta')
                except (ValueError, IndexError):
                    continue
        return self

    def find_crispr_spacers(self, host_file: Path, out_file: Path):
        """This function runs PILER-CR for a given host file and produces output to given file

        Args:
            host_file (Path): Host file to be passed as PILER input.
            out_file (Path): Output file to be created by PILER.
        Raises:
            TypeError: When given obj is of wrong type.
            FileNotFoundError: When given host_file is not file or doesn't exist.
            SubprocessError: When PILER-CR returned error.
        """
        if not isinstance(host_file, Path):
            raise TypeError("Given file is not Path obj.")
        if not isinstance(out_file, Path):
            raise TypeError("Given outfile is not a Path obj.")
        if not host_file.exists():
            raise FileNotFoundError("Host file does not exist.")
        if not host_file.is_file():
            raise FileNotFoundError("Given file is not a file.")

        try:
            subprocess.run(['pilercr', '-in', str(host_file), '-out', str(out_file)], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        except subprocess.CalledProcessError as e:
            raise subprocess.SubprocessError("PilerCR returned error. Check your input", e.output)


def main(args):
    """Main function for running module.
    Args:
        args (argparse obj): Arguments from right subparser.
    """
    print(utils.LOGO)
    args.short_config = utils.parse_config(args.short_config, {
        "task": "blastn-short",
        "num_threads": os.cpu_count(),
        "evalue": 1,
        "gapopen": 10,
        "gapextend": 2,
        "penalty": -1,
        "word_size": 7,
        "dust": "no",
    })

    print("Starting analysis...")
    print("Aggregating files, retrieving crispr spacers...")
    finder = CrisprFinder(Path(args.host_dir), "-")
    finder.retrieve_spacers()

    if args.create_db_name:
        vir_db, vir_db_output = blast.Database(Path(args.virus_dir), args.create_db_name).create()
        print(vir_db_output)
    else:
        vir_db = blast.Database(Path('.'), args.use_db_name)

    print("Quering...")
    query_df = vir_db.query(Path("crispr_spacers/"),
                            config=args.short_config,
                            blast_format="10 qseqid sseqid score qlen length mismatch gaps",
                            headers=("Spacer", "Virus", "Score", "Qlen", "Alen", "Mis", "Gap"))
    if args.clear_after:
        vir_db.clear_files()
    shutil.rmtree(Path("crispr_spacers/"))

    query_df[["Score", "Qlen", "Alen", "Mis", "Gap"]].apply(pd.to_numeric)
    query_df['Allowed'] = query_df['Qlen'] - (
            query_df['Alen'] - query_df['Mis'] - query_df['Gap'])
    query_df['Allowed'] = query_df['Allowed'].apply(pd.to_numeric)
    short_results = query_df.drop(columns=['Qlen', 'Alen', 'Mis', 'Gap'])
    short_results = short_results[short_results['Allowed'] <= args.allowed_mis].drop(columns='Allowed').reset_index(drop=True)

    short_results.rename(columns={"Spacer": "Host"}, inplace=True)
    short_results = short_results.reindex(["Virus", "Host", "Score"], axis=1)
    short_results["Host"] = short_results["Host"].map(lambda x: "|".join(x.split("|")[:2]))
    short_results = short_results.reset_index(drop=True)

    # short_results = short_results.groupby(["Virus", "Spacer"]).sum().reset_index()
    # idx = short_results.groupby(['Virus'])['Score'].idxmax()
    # crispr_results: pd.DataFrame = short_results.loc[idx].sort_values('Score', ascending=False).reset_index(drop=True)

    print("blastn-short results (vir_genome-spacers query): ", short_results, sep='\n')

    short_results.to_csv(args.output_file, index=False)
    print("Saved files to", args.output_file)


class Expando(object):
    pass


if __name__ == '__main__':
    args = Expando()
    args.virus_dir = Path("example_data/virus")
    args.host_dir = Path("example_data/host")
    args.create_db_name = "vir_db"
    args.short_config = ""
    args.clear_after = True
    args.output_file = "crispr_results.csv"
    args.allowed_mis = 1
    main(args)



#!/usr/bin/env python3
from setuptools import setup

with open("README.md") as fh:
    description = fh.read()

setup(
    name='jasper_vh',
    version='1.0.3',
    description="Just a simple virus's hosts predictor",
    long_description=description,
    long_description_content_type='text/markdown',
    author='Milosz Chodkowski',
    keywords='bioinformatics sequence DNA trna CRISPR blast virus host',
    license="GPLv3",
    author_email='milosz.chodkowski@student.put.poznan.pl',
    url="https://github.com/777moneymaker/jasper",
    download_url="https://github.com/777moneymaker/jasper/archive/v1.0.2.tar.gz",
    packages=['jasper'],
    install_requires=[
        "biopython",
        "numpy",
        "pandas",
    ],
    scripts=['jasper-vh'],
    classifiers=[
        "Development Status :: 4 - Beta",
        'Environment :: Console',
        "Intended Audience :: Education",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ]
)



#!/usr/bin/env python3
import os
import subprocess
from pathlib import Path

from Bio import SeqIO
from Bio.Blast.Applications import NcbimakeblastdbCommandline
from jasper import blast
from jasper import utils


class tRNAScanner(blast.Database):
    def __init__(self, source_dir: Path, target_dir: Path, name: str) -> None:
        super(tRNAScanner, self).__init__(source_dir, name)
        self.target_dir = target_dir

    def create(self):
        '''
        TODO: FIX THAT
        :return:
        '''

        try:
            cmd = NcbimakeblastdbCommandline(input_file="blast_input.fasta",
                                             dbtype="nucl",
                                             title=self.name,
                                             out=self.name)
            makeblastdb_output = subprocess.run(str(cmd), capture_output=True)  # shell=True
            if makeblastdb_output.stderr:
                if not Path("blast_input.fasta").exists():
                    raise subprocess.SubprocessError(
                        "Makeblastdb returned error. Input file for Makeblastdb does not exist. Check your input.",
                        str(cmd))
                else:
                    raise subprocess.SubprocessError("Makeblastdb returned error. Check your input.",
                                                     makeblastdb_output.stderr.decode())
        except Exception:
            raise
        finally:
            if Path("blast_input.fasta").exists():
                Path("blast_input.fasta").unlink()
            return self, makeblastdb_output.stdout.decode()

    def scan(self, source_dir: Path, out_file: Path, mode: str = 'general'):
        aggregate = Path("trna_aggregate.fasta")
        for i, source_file in enumerate(source_dir.iterdir(), 1):
            if not source_file.name.endswith(utils.TYPES):
                continue

            with open(aggregate, 'a+') as aggregate_fh:
                aggregate_fh.write(self._repair_fasta(source_file))
            print(f"Performed {i} files aggregation for tRNA-scan", end='\r')

        print("\nRunning tRNAscan-SE")
        trna_output = self.run_trnascan(aggregate, out_file, os.cpu_count(), mode)
        print(trna_output.stdout.decode())
        aggregate.unlink()

        with open(out_file, 'r') as trna_fh:
            trnas = [(str(record.seq), record.name) for record in SeqIO.parse(trna_fh, 'fasta')]
            out_file.unlink()
            return trnas

    def run_trnascan(self, file: Path, out_file: Path, num_threads: int, mode: str):
        if mode == 'bacterial':
            mode = '-B'
        elif mode == 'pro':
            mode = "-P"
        else:
            mode = "-G"

        res = subprocess.run(['tRNAscan-SE', str(file), '-a', str(out_file), '--thread', str(num_threads), mode], capture_output=True)
        return res


if __name__ == '__main__':
    t = tRNAScanner(source_dir=Path("example_data/host"), target_dir=Path("example_data/virus"), name="trna_db")
    #host_trnas = t.scan(t.source_dir, Path("host_trnas.fasta"))
    vir_trnas = t.scan(t.target_dir, Path("virus_trnas.fasta"), mode='general')
    print(*vir_trnas, sep='\n')
